/*---------------------------------------------------------------------------------------------------------
 * uart-bootloader.c
 *
 * Copyright (c) 2011 Robert Meyer, Frank Meyer
 *
 * ATMEGA168 @ 8 MHz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *---------------------------------------------------------------------------------------------------------
 */

#define F_CPU               8000000L               // system clock in Hz, e.g. 4000000

#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/boot.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <inttypes.h>

int             main (void);
static void     boot_update_flash (void);
static void     boot_uart_init (void);
static uint8_t  boot_uart_getc (void);
static void     boot_uart_putc (unsigned char);
static void     boot_uart_puts (char *);
static uint8_t  boot_hex2toi (char *);

#define boot_program_page_erase(page)               \
{                                                   \
    eeprom_busy_wait ();                            \
    boot_page_erase ((uint32_t) page);              \
    boot_spm_busy_wait ();                          \
}

#define boot_program_page_write(page)               \
{                                                   \
    boot_page_write ((uint32_t) page);              \
    boot_spm_busy_wait ();                          \
    boot_rww_enable ();                             \
}


#if 1

#define LED_1_PORT          PORTD
#define LED_1_DDR           DDRD
#define LED_1               6

#define LED_2_PORT          PORTD
#define LED_2_DDR           DDRD
#define LED_2               7

#define LED_3_PORT          PORTB
#define LED_3_DDR           DDRB
#define LED_3               0

#define LED_1_ON            LED_1_PORT |= 1<<LED_1
#define LED_1_OFF           LED_1_PORT &= ~(1<<LED_1)
#define LED_2_ON            LED_2_PORT |= 1<<LED_2
#define LED_2_OFF           LED_2_PORT &= ~(1<<LED_2)
#define LED_3_ON            LED_3_PORT |= 1<<LED_3
#define LED_3_OFF           LED_3_PORT &= ~(1<<LED_3)

static void
led_init (void)
{
    LED_1_OFF;
    LED_1_DDR |= (1<<LED_1);

    LED_2_OFF;
    LED_2_DDR |= (1<<LED_2);

    LED_3_OFF;
    LED_3_DDR |= (1<<LED_3);
}

#endif // 1

//-------------------------------------

int
main (void)
{
    void        (*funcptr) (void) = 0x0000;
    uint8_t     temp;

    wdt_disable ();

#if 1
    led_init();
    LED_1_ON;
    LED_2_ON;
    LED_3_ON;
#endif
    
    boot_uart_init ();
    _delay_ms(3000);

    if (UCSR0A & (1<<RXC0))                         // character available?
    {                                               // yes...
        temp = boot_uart_getc ();

        if (temp == '$')
        {
            boot_uart_putc (temp);
            boot_update_flash ();
        }
    }

#if 1
    LED_1_OFF;
    LED_2_OFF;
    LED_3_OFF;
#endif

    _delay_ms(1000);
 
    // reset interrupt vectors
    temp = MCUCR;
    MCUCR = temp | (1<<IVCE);
    MCUCR = temp & ~(1<<IVSEL);

    funcptr();                                          // jump to application section
    /* NOTREACHED */
    return 0;
}

#define BOOT_CMD_BUFLEN     16

#define F_INTERRUPTS        1000                                                // 1000 interrupts per second
static volatile uint8_t     seccnt;

void
boot_update_flash (void)
{
    int             pageoffset = 0;
    char            uart_buf[64];
    uint8_t         len;
    uint8_t         ch;
    uint8_t         idx;
    uint8_t         datalen;
    unsigned int    address;
    unsigned int    start_address = 0x0000;
    uint8_t         addrh;
    uint8_t         addrl;
    uint8_t         rectype = 0x00;
    char *          dataptr;
    uint8_t         chcksum;
    int             sum = 0;
    uint8_t         line_ok;

    for (;;)
    {
        len = 0;

        while (len < 64 && (ch = boot_uart_getc()) != '\n')                           
        {
            if (ch != '\0' && ch != '\r')       // 1st character is always a '\0', ignore it!
            {
                uart_buf[len++] = ch;
            }
        }

        uart_buf[len] = '\0';

        line_ok = 0;                        // 0 = not OK

        if (uart_buf[0] == ':' && len >= 11)
        {
            datalen = boot_hex2toi (uart_buf + 1);
            addrh   = boot_hex2toi (uart_buf + 3);
            addrl   = boot_hex2toi (uart_buf + 5);
            rectype = boot_hex2toi (uart_buf + 7);

            address = addrh << 8 | addrl;
            dataptr = uart_buf + 9;

            sum = datalen + addrh + addrl + rectype;

            if (pageoffset == 0)
            {
                start_address = address;

                if (rectype != 0x01)
                {
                    boot_program_page_erase (start_address);
                }
            }
            else if (rectype == 0x00)   // be sure, that pageoffset points to correct offset if error in transmission occured
            {
                pageoffset = address - start_address;

                if (pageoffset < 0 || pageoffset > SPM_PAGESIZE - datalen)  // pageoffset too large?
                {
                    len = 0;                                                // yes, indicate error
                }
            }
            else    // never should be! pageoffset > 0 and rectype == 0x01
            {
                // pageoffset = 0;
            }

            if (len == 9 + 2 * datalen + 2)
            {
                for (idx = 0; idx < datalen; idx++)
                {
                    sum += boot_hex2toi (dataptr);
                    dataptr +=2;
                }

                sum = (0x100 - (sum & 0xff)) & 0xff;
                chcksum = boot_hex2toi (dataptr);

                if (sum == chcksum)
                {
                    dataptr = uart_buf + 9;

                    for (idx = 0; idx < datalen; idx++)
                    {
                        if ((idx & 0x01) == 0)                                  // even?
                        {
                            uint16_t w = boot_hex2toi (dataptr);
                            w += (boot_hex2toi (dataptr + 2) << 8);
                            boot_page_fill ((uint32_t) (start_address + pageoffset), w);
                        }

                        pageoffset++;
                        dataptr +=2;
                    }

                    if (address < 0x3800)       // only flash application section
                    {
                        line_ok = 1;

                        if (pageoffset >= SPM_PAGESIZE)
                        {
                            boot_program_page_write ((uint32_t) start_address);
                            pageoffset = 0;
                            line_ok = 3;
                        }
                    }
                    else
                    {
                        line_ok = 2;            // ignore boot loader section
                    }
                }
                else
                {
                    pageoffset -= datalen;
                }
            }
        }

        if (line_ok > 0 && rectype == 0x01)     // EOF
        {
            if (pageoffset > 0)
            {
                boot_program_page_write ((uint32_t) start_address);

                line_ok = 3;
            }

        }

        uart_buf[0] = 'O';
        uart_buf[1] = 'K';
        uart_buf[2] = line_ok + '0';           // 1 = OK
        uart_buf[3] = '\r';
        uart_buf[4] = '\n';
        uart_buf[5] = '\0';
        boot_uart_puts (uart_buf);

        if (line_ok != 0 && rectype == 0x01)    // EOF
        {
            break;
        }
    }
    return;
}

#define BAUD                                    9600L
#include <util/setbaud.h>

#ifdef UBRR0H

#define UART0_UBRRH                             UBRR0H
#define UART0_UBRRL                             UBRR0L
#define UART0_UCSRA                             UCSR0A
#define UART0_UCSRB                             UCSR0B
#define UART0_UCSRC                             UCSR0C
#define UART0_UDRE_BIT_VALUE                    (1<<UDRE0)
#define UART0_UCSZ1_BIT_VALUE                   (1<<UCSZ01)
#define UART0_UCSZ0_BIT_VALUE                   (1<<UCSZ00)
#ifdef URSEL0
#define UART0_URSEL_BIT_VALUE                   (1<<URSEL0)
#else
#define UART0_URSEL_BIT_VALUE                   (0)
#endif
#define UART0_TXEN_BIT_VALUE                    (1<<TXEN0)
#define UART0_RXEN_BIT_VALUE                    (1<<RXEN0)
#define UART0_UDR                               UDR0
#define UART0_U2X                               U2X0
#define UART0_RXC                               RXC0

#else

#define UART0_UBRRH                             UBRRH
#define UART0_UBRRL                             UBRRL
#define UART0_UCSRA                             UCSRA
#define UART0_UCSRB                             UCSRB
#define UART0_UCSRC                             UCSRC
#define UART0_UDRE_BIT_VALUE                    (1<<UDRE)
#define UART0_UCSZ1_BIT_VALUE                   (1<<UCSZ1)
#define UART0_UCSZ0_BIT_VALUE                   (1<<UCSZ0)
#ifdef URSEL
#define UART0_URSEL_BIT_VALUE                   (1<<URSEL)
#else
#define UART0_URSEL_BIT_VALUE                   (0)
#endif
#define UART0_TXEN_BIT_VALUE                    (1<<TXEN)
#define UART0_RXEN_BIT_VALUE                    (1<<RXEN)
#define UART0_UDR                               UDR
#define UART0_U2X                               U2X
#define UART0_RXC                               RXC

#endif

static void
boot_uart_init (void)
{
    UART0_UBRRH = UBRRH_VALUE;                                                                      // set baud rate
    UART0_UBRRL = UBRRL_VALUE;

#if USE_2X
    UART0_UCSRA |= (1<<UART0_U2X);
#else
    UART0_UCSRA &= ~(1<<UART0_U2X);
#endif

    UART0_UCSRC = UART0_UCSZ1_BIT_VALUE | UART0_UCSZ0_BIT_VALUE | UART0_URSEL_BIT_VALUE;
    UART0_UCSRB |= UART0_TXEN_BIT_VALUE;                                                            // enable UART TX
    UART0_UCSRB |= UART0_RXEN_BIT_VALUE;                                                            // activate UART RX
}

static uint8_t
boot_uart_getc (void)
{
    while (!(UCSR0A & (1<<RXC0)))                   // character available?
    {                                               // no...
        ;                                           // wait
    }

    return UDR0;                                    // return character from register UDR0
}

static void
boot_uart_putc (unsigned char ch)
{
    while (!(UCSR0A & (1<<UDRE0)))                  // wait for ready to send
    {
        ;
    }

    UDR0 = ch;                                  // write character to uart
}

static void
boot_uart_puts (char * s)
{
    unsigned char * us = (unsigned char *) s;

    while (*us)
    {
        boot_uart_putc (*us);
        us++;
    }
    return;
}

static uint8_t
boot_hex2toi (char * p)
{
    uint8_t i = 0;
    uint8_t ch;
    uint8_t idx;

    for (idx = 0; idx < 2; idx++)
    {
        ch = *p++;

        if (ch >= '0' && ch <= '9')
        {
            i |= (ch - '0');
        }
        else if (ch >= 'A' && ch <= 'F')
        {
            i |= (ch - 'A' + 10);
        }
        else if (*p >= 'a' && *p <= 'f')
        {
            i |= (ch - 'a' + 10);
        }

        if (idx < 1)
        {
            i <<= 4;
        }
    }

    return (i);
}


