/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * @file uart-flash.c
 *
 * Copyright (c) 2011 Robert & Frank Meyer - frank(at)fli4l.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef unix
#define HANDLE  int
#include <stdint.h>
#else
#include <windows.h>
typedef unsigned char       uint8_t;
#endif

#define TRUE                1
#define FALSE               0
#define ERR                 (-1)
#define OK                  0
#define WARN                1

#define TIMEOUT_SEC         5               // timeout after 5 seconds

static HANDLE               uartfd;
static char *               com;

#define SPM_PAGESIZE        128
#define LINES_PER_PAGE      8
#define RCVBUFLEN           64

char *                      pgm_name;
static unsigned char        flash[16384];
static unsigned int         address_low     = 0xffff;
static unsigned int         address_high    = 0x0000;

#ifndef unix
void
win_perror (LPTSTR lpszFunction) 
{ 
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL,
                  dw,
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  (LPTSTR) &lpMsgBuf,
                  0, NULL );

    fprintf (stderr, "%s: error %d: %s", lpszFunction, dw, lpMsgBuf); 
    LocalFree(lpMsgBuf);
}
#endif

/*------------------------------------------------------------------------------------------------------------------------
 * uart_open ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static HANDLE
uart_open (char * device, int baudrate)
{
    HANDLE hPort;

#ifdef unix
    hPort = open (device, O_RDWR);

    if (hPort < 0)
    {
        perrror (com);
        exit (1);
    }

#else

    hPort = CreateFile (device,                         // Pointer to the name of the port
                        GENERIC_READ | GENERIC_WRITE,   // Access (read-write) mode
                        0,                              // Share mode
                        NULL,                           // Pointer to the security attribute
                        OPEN_EXISTING,                  // How to open the serial port
                        0,                              // Port attributes
                        NULL);                          // HANDLE to port with attribute

    if (hPort != INVALID_HANDLE_VALUE)
    {
        DCB PortDCB;

        // Initialize the DCBlength member.
        PortDCB.DCBlength = sizeof (DCB);

        // Get the default port setting information.
        GetCommState (hPort, &PortDCB);

        // Change the DCB structure settings.
        PortDCB.BaudRate            = baudrate;             // Current baud
        PortDCB.fBinary             = TRUE;                 // Binary mode; no EOF check
        PortDCB.fParity             = TRUE;                 // Enable parity checking
        PortDCB.fOutxCtsFlow        = FALSE;                // No CTS output flow control
        PortDCB.fOutxDsrFlow        = FALSE;                // No DSR output flow control
        PortDCB.fDtrControl         = DTR_CONTROL_ENABLE;   // DTR flow control type
        PortDCB.fDsrSensitivity     = FALSE;                // DSR sensitivity
        PortDCB.fTXContinueOnXoff   = TRUE;                 // XOFF continues Tx
        PortDCB.fOutX               = FALSE;                // No XON/XOFF out flow control
        PortDCB.fInX                = FALSE;                // No XON/XOFF in flow control
        PortDCB.fErrorChar          = FALSE;                // Disable error replacement
        PortDCB.fNull               = FALSE;                // Disable null stripping
        PortDCB.fRtsControl         = RTS_CONTROL_ENABLE;   // RTS flow control
        PortDCB.fAbortOnError       = FALSE;                // Do not abort reads/writes on error
        PortDCB.ByteSize            = 8;                    // Number of bits/byte, 4-8
        PortDCB.Parity              = NOPARITY;             // 0-4=no,odd,even,mark,space
        PortDCB.StopBits            = ONESTOPBIT;           // 0,1,2 = 1, 1.5, 2

        // Configure the port according to the specifications of the DCB
        // structure.
        if (!SetCommState (hPort, &PortDCB))
        {
            win_perror ("SetCommState");                    // Could not configure the serial port.
            CloseHandle (hPort);
            exit (1);
        }
    }
    else
    {
        win_perror (device);
        exit (1);
    }
#endif
    return (hPort);
}

void
uart_set_timeout (HANDLE hPort, int msec)
{
    COMMTIMEOUTS CommTimeouts;
    // Retrieve the time-out parameters for all read and write operations
    // on the port.
    GetCommTimeouts (hPort, &CommTimeouts);

    // Change the COMMTIMEOUTS structure settings.
    CommTimeouts.ReadIntervalTimeout = MAXDWORD;
    CommTimeouts.ReadTotalTimeoutMultiplier = 0;
    CommTimeouts.ReadTotalTimeoutConstant = msec;
    CommTimeouts.WriteTotalTimeoutMultiplier = 10;
    CommTimeouts.WriteTotalTimeoutConstant = 1000;

    // Set the time-out parameters for all read and write operations
    // on the port.
    if (!SetCommTimeouts (hPort, &CommTimeouts))
    {
        // Could not set the time-out parameters.
        win_perror ("SetCommTimeouts");
        CloseHandle(hPort);
        exit (1);
    }
}
                  
/*------------------------------------------------------------------------------------------------------------------------
 * uart_poll ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static int
uart_poll (HANDLE fd)
{
    char    buf[1];
    int     ch = -2;
    int     idx;
    int     rtc;

    uart_set_timeout (fd, 1000);

    for (idx = 0; idx < TIMEOUT_SEC; idx++)
    {
#ifdef unix
        if ((rtc = read (fd, buf, 1)) >= 0)
#else
        if (ReadFile (fd, buf, 1, &rtc, NULL))
#endif
        {
            if (rtc == 1)
            {
                ch = buf[0];
                break;
            }
        }
        else
        {
            ch = -1;
            break;
        }
    }

    uart_set_timeout (fd, 0);
    return (ch);
}

/*------------------------------------------------------------------------------------------------------------------------
 * uart_read ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static int
uart_read (HANDLE fd, char * buf, int len)
{
    int            rtc;
    int            offset  = 0;

    for (;;)
    {
#ifdef unix
        if ((rtc = read (fd, buf + offset, len)) < 0)
#else
        if (! ReadFile (fd, buf + offset, len, &rtc, NULL))
#endif
        {
            return (-1);                // error occured
        }
        else
        {
            offset += rtc;
            len -= rtc;

            if (*(buf + offset - 1) == '\n')
            {
                break;
            }
        }
    }

    return (offset);
}

/*------------------------------------------------------------------------------------------------------------------------
 * uart_write ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static int
uart_write (HANDLE fd, char * buf, int len)
{
    int rtc;

#ifdef unix
    if ((rtc = write (fd, buf, len)) > 0)
#else
    if (WriteFile (fd, buf, len, &rtc, NULL))
#endif
    {
        return rtc;
    }
    return (-1);
}

/*------------------------------------------------------------------------------------------------------------------------
 * uart_close ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static void
uart_close (HANDLE fd)
{
#ifdef unix
    (void) close (fd);
#else
    (void) CloseHandle (fd);
#endif
}

#define BUFLEN        512

/*------------------------------------------------------------------------------------------------------------------------
 * uart_readline ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static int
uart_readline (HANDLE fd, char * buf, int rcvlen, int *lenp)
{
    int len;

    len = uart_read (fd, buf, rcvlen);

    if (len < 0)
    {
        fprintf (stderr, "read from %s failed. trying to reconnect...\n", com);
        return (ERR);
    }

    buf[len] = '\0';

    if (len == 0)
    {
        fprintf (stderr, "got no answer from %s\n", com);
        return (WARN);
    }

    if (buf[0] != 'O' || buf[1] != 'K')
    {
        fprintf (stderr, "got invalid answer from %s '%s'.\n", com, buf);
        return (WARN);
    }

    *lenp = len;
    return (OK);
}

/*------------------------------------------------------------------------------------------------------------------------
 * uart_send_command ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static int
uart_send_command (HANDLE fd, const char * sndbuf, char * rcvbuf, int rcvlen)
{
    static char    buf[BUFLEN];
    char *    bufp;
    int        retry;
    int        len;
    int        rtc;

    for (retry = 0; retry < 2; retry++)
    {
        sprintf (buf, "%s\r\n", sndbuf);
        len = strlen (buf);

        rtc = uart_write (fd, buf, len);

        if (rtc < 0)
        {
            fprintf (stderr, "write failed\n");
            return (-1);
        }

        if (rtc != len)
        {
            fprintf (stderr, "write to %s failed.", com);
            return (WARN);
        }

        rtc = uart_readline (fd, buf, rcvlen, &len);

        if (rtc == OK)
        {
            break;
        }
    }

    if (rtc != OK)
    {
        return (rtc);
    }

    if (len >= 2 && buf[len - 2] == '\r' && buf[len - 1] == '\n')
    {
        buf[len - 2] = '\0';
    }
    else
    {
        fprintf (stderr, "missing CRLF in answer from %s: '%s'.\n", com, buf);
        return (WARN);
    }

    bufp = buf;
    strcpy (rcvbuf, bufp);
    return (OK);
}

/*------------------------------------------------------------------------------------------------------------------------
 * usage ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static void
usage (char * pgm)
{
    fprintf (stderr, "usage: %s COMx: hex-file\n", pgm);
    exit (1);
}

/*------------------------------------------------------------------------------------------------------------------------
 * hex2toi ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static uint8_t
hex2toi (char * p)
{
    uint8_t i = 0;
    uint8_t ch;
    uint8_t idx;

    for (idx = 0; idx < 2; idx++)
    {
        ch = *p++;

        if (ch >= '0' && ch <= '9')
        {
            i |= (ch - '0');
        }
        else if (ch >= 'A' && ch <= 'F')
        {
            i |= (ch - 'A' + 10);
        }
        else if (*p >= 'a' && *p <= 'f')
        {
            i |= (ch - 'a' + 10);
        }

        if (idx < 1)
        {
            i <<= 4;
        }
    }

    return (i);
}

/*------------------------------------------------------------------------------------------------------------------------
 * check_hex_file ()
 *------------------------------------------------------------------------------------------------------------------------
 */
int
check_hex_file (FILE * fp)
{
    char            buf[64];
    int             line_ok;
    unsigned char   len;
    unsigned char   ch;
    unsigned char   idx;
    unsigned char   datalen;
    unsigned int    address;
    unsigned char   addrh;
    unsigned char   addrl;
    unsigned char   rectype;
    char *          dataptr;
    unsigned char   chcksum;
    int             sum = 0;
    int             do_flash;

    while (fgets (buf, 64, fp))
    {
        line_ok = FALSE;

        len = strlen (buf);

        if (len > 2 && *(buf + len - 2) == '\r' && *(buf + len - 1) == '\n')
        {
            *(buf + len - 2) = '\0';
        }

        len = strlen (buf);

        if (buf[0] == ':' && len >= 11)
        {
            datalen = hex2toi (buf +  1);
            addrh   = hex2toi (buf +  3);
            addrl   = hex2toi (buf +  5);
            address = addrh << 8 | addrl;
            rectype = hex2toi (buf + 7);
            dataptr = buf + 9;

            sum = datalen + addrh + addrl + rectype;

            if (len == 9 + 2 * datalen + 2)
            {
                if (address < 0x3800)
                {
                    if (address_low > address)
                    {
                        address_low = address;
                    }
                }

                for (idx = 0; idx < datalen; idx++)
                {
                    ch = hex2toi (dataptr);

                    if (address + idx < 0x3800)
                    {
                        do_flash = TRUE;
                    }
                    else
                    {
                        do_flash = FALSE;
                    }

                    if (do_flash)
                    {
                        flash[address + idx] = ch;

                        if (address_high < address + idx)
                        {
                            address_high = address + idx;
                        }
                    }

                    sum += ch;
                    dataptr +=2;
                }

                sum = (0x100 - (sum & 0xff)) & 0xff;
                chcksum = hex2toi (dataptr);

                if (sum == chcksum)
                {
                    line_ok = TRUE;
                }
                else
                {
                    fprintf (stderr, "invalid checksum: sum: 0x%02x chcksum: 0x%02x\n", sum, chcksum);
                }
            }
            else
            {
                fprintf (stderr, "invalid len: %d, expected %d\n", len, 12 + 2 * datalen + 2);
            }
        }
        else
        {
            fprintf (stderr, "invalid format, len: %d\n", len);
        }

        if (! line_ok)
        {
            fprintf (stderr, "line '%s' not ok\n", buf);
            return (ERR);
        }
    }

    return (OK);
}

/*------------------------------------------------------------------------------------------------------------------------
 * send_line ()
 *------------------------------------------------------------------------------------------------------------------------
 */
static void
send_line (HANDLE fd, unsigned char * sndbuf)
{
    char    rcvbuf[RCVBUFLEN];
    int     rtc;

    do
    {
        rtc = uart_send_command (fd, sndbuf, rcvbuf, RCVBUFLEN);

        if (rtc == OK)
        {
            if (rcvbuf[0] != 'O' || rcvbuf[1] != 'K')
            {
                rtc = ERR;
            }
        }

        if (rtc == OK)
        {
            if (rcvbuf[2] == '1')
            {
                putchar ('.');
            }
            else if (rcvbuf[2] == '2')
            {
                putchar ('-');
            }
            else if (rcvbuf[2] == '3')
            {
                putchar ('+');
            }
            else
            {
                putchar ('?');
            }
        }
        else
        {
            putchar ('%');
        }
        fflush (stdout);
    } while (rtc != OK);
}

/*------------------------------------------------------------------------------------------------------------------------
 * main ()
 *------------------------------------------------------------------------------------------------------------------------
 */
int
main (int argc, char ** argv)
{
    char            buf[64];
    unsigned char   idx;
    unsigned char   datalen;
    unsigned int    address;
    unsigned char   rectype;
    char            send_ch;
    char            ch;
    int             len;
    int             page;
    int             page_low;
    int             page_high;
    int             line;
    int             rtc = 0;
    int             sum = 0;
    char *          hex_file;
    FILE *          fp;
    HANDLE          fd;

    pgm_name = argv[0];

    if (argc != 3)
    {
        usage (pgm_name);
    }

    com         = argv[1];
    hex_file    = argv[2];

    fp = fopen (hex_file, "rb");

    if (! fp)
    {
        perror (hex_file);
        exit (1);
    }

    if (check_hex_file (fp) != OK)
    {
        fclose (fp);
        exit (1);
    }

    page_low    = address_low / SPM_PAGESIZE;
    page_high   = address_high / SPM_PAGESIZE + 1;

    rewind (fp);

    fd = uart_open (com, 9600);

    printf ("Entering flash mode ... ");
    fflush (stdout);

    send_ch = '$';

    do
    {
        uart_write (fd, &send_ch, 1);
        ch = uart_poll (fd);
        if (ch == '+')
        {
            printf ("flashing from application ... ");
            fflush (stdout);
        }
    } while (ch == '+');

    if (ch == '$')
    {
        printf ("\nStarting Transfer - do not interrupt!\n");

        for (page = page_low; page < page_high; page++)
        {
            for (line = 0; line < LINES_PER_PAGE; line++)
            {
                sum = 0;

                datalen = SPM_PAGESIZE / LINES_PER_PAGE;
                address = page * SPM_PAGESIZE + line * datalen;
                rectype = 0x00;
                sprintf (buf, ":%02X%04X%02X", datalen, address, rectype);

                sum = datalen + (address >> 8) + (address & 0xFF) + rectype;

                len = strlen (buf);

                for (idx = 0; idx < datalen; idx++)
                {
                    sprintf (buf + len + 2 * idx, "%02X", flash[address + idx]);
                    sum += flash[address + idx];
                }

                sum = (0x100 - (sum & 0xff)) & 0xff;
                sprintf (buf + len + 2 * datalen, "%02X", sum);

                send_line (fd, buf);
            }
        }
        send_line (fd, ":00000001FF");        // EOF

        printf ("\nTransfer complete\n");
    }
    else
    {
        if (ch == -2)
        {
            fprintf (stderr, "Timeout reached. Exit\n");
        }
        else
        {
            printf ("Read failed. Exit\n");
        }
        rtc = 1; 
    }

    uart_close (fd);
    fclose (fp);
    return (rtc);
}
